from hitchserve.hitch_exception import WaitingForLogMessageTimeout
from hitchserve.utils import log, warn
import multiprocessing
import functools
import signal
import inspect
import pyuv
import json
import time
import os
import re

# TODO: All of this could do with some refactoring.


class Tail(object):
    def __init__(self, sublog):
        self._sublog = sublog
        self._logfilename = sublog._logfilename
        self.titles = sublog.titles
        self._end_of_file = os.stat(self._logfilename).st_size
        self.max_length_of_titles = max([len(title) for title in self.titles])
        self._closing = False
        self.trigger_exception = False

    def _timer_callback(self, t_handle):
        if time.time() - self._start_time > float(self.timeout):
            if self.raise_exception:
                self.trigger_exception = True
                self._close_handles()

    def _signal_callback(self, handle, signum):
        """Shutdown tail if ctrl-C is received."""
        if signum == signal.SIGINT:
            self._close_handles()

    def _close_handles(self):
        self._closing = True
        if self._timer_handle is not None and not self._timer_handle.closed:
            self._timer_handle.close()
            self._timer_handle = None
        if not self._event_handle.closed:
            self._event_handle.close()
            self._event_handle = None
        if not self._signal_handle.closed:
            self._signal_handle.close()
            self._signal_handle = None

    def _read_callback(self, match_function, do_function, handle, filename, events, error):
        with open(self._logfilename, "r") as filehandle:
            filehandle.seek(self._end_of_file)
            tailportion = filehandle.read()
        self._end_of_file = self._end_of_file + len(tailportion)

        for line in tailportion.split('\n'):
            matching_line = self._sublog._match_service(line)
            if matching_line is not None:
                if match_function(matching_line):
                    do_function(matching_line)

    def _printtuple(self, line):
        if type(line) is tuple:
            log("[{}] {}\n".format(line[0].rjust(self.max_length_of_titles), line[1]))
        else:
            log("{}\n".format(line))

    def wait_and_do(self, match_function, do_function, lines_back=0, timeout=None, raise_exception=True):
        self.loop = pyuv.Loop.default_loop()
        self.timeout = timeout
        self._start_time = time.time()
        self._started = False
        self._returnval = None
        self.raise_exception = raise_exception

        self._signal_handle = pyuv.Signal(self.loop)
        self._signal_handle.start(self._signal_callback, signal.SIGINT)

        if timeout is None:
            self._timer_handle = None
        else:
            self._timer_handle = pyuv.Timer(self.loop)
            self._timer_handle.start(self._timer_callback, 0.1, 0.1)

        self._event_handle = pyuv.fs.FSEvent(self.loop)

        self._event_handle.start(self._logfilename, 0,
            functools.partial(
                self._read_callback,
                match_function,
                do_function,
            )
        )

        lines = self._sublog.lines()

        for line in lines[len(lines) - lines_back:]:
            if match_function(line):
                do_function(line)

        if not self._started:
            self._started = True
            self.loop.run()
            self.loop = None
        else:
            self.loop = None

        if self.trigger_exception:
            raise WaitingForLogMessageTimeout("Timeout waiting for log line.")
        return self._returnval

    def follow(self, lines_back=0):
        return self.wait_and_do(
            lambda line: True,
            lambda line: self._printtuple(line),
            lines_back=lines_back,
            timeout=None,
            raise_exception=False
        )

    def grep(self, function, lines_back, timeout=None, raise_exception=False):
        return self.wait_and_do(
            lambda line: function,
            lambda line: self._printtuple(line),
            lines_back=lines_back,
            timeout=timeout,
            raise_exception=raise_exception
        )


    def until(self, function, lines_back=0, timeout=None, raise_exception=True):
        def on_match(line):
            self._returnval = line[1]
            self._close_handles()

        return self.wait_and_do(function, on_match, lines_back=lines_back, timeout=timeout, raise_exception=raise_exception)


    def until_json(self, function, lines_back=0, timeout=None, raise_exception=True):
        def json_match(line):
            try:
                return function(json.loads(line[1]))
            except ValueError:
                return False
            except IndexError:
                return False

        def on_match(line):
            self._returnval = json.loads(line[1])
            self._close_handles()

        return self.wait_and_do(json_match, on_match, lines_back=lines_back, timeout=timeout, raise_exception=raise_exception)

    def lines(self):
        lines = []
        with open(self._logfilename, "r") as log_handle:
            for line in log_handle:
                matching_line = self._sublog._match_service(line)
                if matching_line is not None:
                    lines.append(matching_line)
        return lines

    def cat(self, numlines=None):
        """Return a list of lines output by this service."""
        if len(self.titles) == 1:
            lines = self.lines()
            if numlines is not None:
                lines = lines[len(lines)-numlines:]
            log("\n".join(lines))
        else:
            lines = [self._printtuple(line[0], line[1]) for line in self.lines()]
            if numlines is not None:
                lines = lines[len(lines)-numlines:]
            log("".join(lines))


class SubLog(object):
    def __init__(self, name, titles):
        self.titles = titles
        self.name = name
        self._end_of_file = 0
        self.max_length_of_titles = max([len(title) for title in titles])

    def _match_service(self, line_with_color):
        """Return line if line matches this service's name, return None otherwise."""
        line = re.compile("(\x1b\[\d+m)+").sub("", line_with_color)        # Strip color codes

        regexp = re.compile(r"^\[(.*?)\]\s(.*?)$")
        if regexp.match(line):
            title = regexp.match(line).group(1).strip()
            if title in self.titles:
                return (title, regexp.match(line).group(2))
        return None

    @property
    def tail(self):
        """Get a Tail object for these logs."""
        return Tail(self)

    def lines(self):
        """Return a list of lines output by this service."""
        lines = []
        with open(self._logfilename, "r") as log_handle:
            for line in log_handle:
                matching_line = self._match_service(line)
                if matching_line is not None:
                    lines.append(matching_line)
        return lines

    def __repr__(self):
        """Generate a string representation."""
        return "\n".join(["[{}] {}".format(title.rjust(self.max_length_of_titles), line) for title, line in self.lines()])

    def __str__(self):
        return self.__repr__()

    def json(self):
        """Return a list of JSON objects output by this service."""
        lines = []
        for line in self.lines():
            try:
                if len(line) == 1:
                    lines.append(json.loads(line, strict=False))
                else:
                    lines.append(json.loads(line[1], strict=False))
            except ValueError:
                pass
        return lines

class ServiceLogs(SubLog):
    """Service log handling - tailing, matching lines, extracting JSON."""

    def __init__(self, name):
        self._name = name
        sout = name
        serr = "Err " + name
        setupout = "Setup " + name
        setuperr = "Err Setup " + name
        postout = "Post " + name
        posterr = "Post Err " + name

        self.out = SubLog(name, [sout ])
        self.err = SubLog(name, [serr ])
        self.setup = SubLog(name, [setupout, setuperr, ])
        self.setup.out = SubLog(name, [setupout, ])
        self.setup.err = SubLog(name, [setuperr, ])
        self.poststart = SubLog(name, [postout, posterr,])
        self.poststart.out = SubLog(name, [postout, ])
        self.poststart.err = SubLog(name, [posterr, ])
        super(ServiceLogs, self).__init__(name, titles=[sout, serr, setupout, setuperr, postout, posterr, ])

    def set_logfilename(self, logfilename):
        self._logfilename = logfilename
        self.out._logfilename = logfilename
        self.err._logfilename = logfilename
        self.setup._logfilename = logfilename
        self.setup.out._logfilename = logfilename
        self.setup.err._logfilename = logfilename
        self.poststart._logfilename = logfilename
        self.poststart.out._logfilename = logfilename
        self.poststart.err._logfilename = logfilename



from os import sep, path, makedirs, system, remove, listdir
import shutil
import sys


class HitchDir(object):
    def __init__(self, project_directory, hitch_dir=None):
        self.project_directory = project_directory

        # 3 directories above virtualenv python binary is the hitchdir
        if hitch_dir is None:
            self.hitch_dir = path.abspath(
                path.join(path.dirname(sys.executable), "..", "..")
            )
        else:
            self.hitch_dir = hitch_dir

    def clean(self):
        if not path.exists(self.hitch_dir):
            makedirs(self.hitch_dir)
            makedirs(self.hitch_dir)
        if path.exists(self.faketime()):
            remove(self.faketime())
        shutil.rmtree(self.stdlogdir(), ignore_errors=True)
        makedirs(self.stdlogdir())
        if path.exists(self.testlog()):
            remove(self.testlog())
        self.remove_run_dir()
        makedirs(self.rundir())

    def save_pgid(self, name, pgid):
        with open(self.rundir() + sep + name + '.pgid', "w") as pgid_file_handle:
            pgid_file_handle.write(str(pgid))

    def save_pid(self, pid):
        with open(self.rundir() + sep + 'driver_process.pid', "w") as pid_file_handle:
            pid_file_handle.write(str(pid))

    def remove_run_dir(self):
        if path.exists(self.rundir()):
            shutil.rmtree(self.rundir())

    def old_pid(self):
        if path.exists(path.join(self.rundir(), 'driver_process.pid')):
            with open(path.join(self.rundir(), 'driver_process.pid')) as pid_file_handle:
                return int(pid_file_handle.read())
        else:
            return None

    def old_pgids(self):
        pgids = []
        if path.exists(self.rundir()):
            for filename in listdir(self.rundir()):
                with open(path.join(self.rundir(), filename)) as pgid_file_handle:
                    pgids.append(int(pgid_file_handle.read()))
        return pgids

    def rundir(self):
        return self.hitch_dir + sep + "run"

    def testlog(self):
        return self.hitch_dir + sep + "test.log"

    def stdlogdir(self):
        return self.hitch_dir + sep + "stdlog"

    def faketime(self):
        return self.hitch_dir + sep + "faketime.txt"

    def pgid(self):
        return self.hitch_dir + sep + "hitch.pgid"

    def driverout(self):
        return self.stdlogdir() + sep + "driver.out"

    def drivererr(self):
        return self.stdlogdir() + sep + "driver.err"

    def setup_out(self, name):
        return self.stdlogdir() + sep + name.lower() + "_setup.out"

    def setup_err(self, name):
        return self.stdlogdir() + sep + name.lower() + "_setup.err"

    def poststart(self, name):
        return self.stdlogdir() + sep + name.lower() + "_poststart.out"

    def poststart_err(self, name):
        return self.stdlogdir() + sep + name.lower() + "_poststart.err"



class HitchException(Exception):
    pass

class HitchServeException(Exception):
    pass

class ServiceStartupTimeoutException(HitchServeException):
    pass

class ServiceSuddenStopException(HitchServeException):
    pass

class ServiceMisconfiguration(HitchServeException):
    pass

class BundleMisconfiguration(HitchServeException):
    pass

class WaitingForLogMessageTimeout(HitchServeException):
    pass



from hitchserve.service_handle import ServiceHandle
from hitchserve.hitch_exception import ServiceSuddenStopException
from colorama import Fore
import time


class ServiceEngine(object):
    def __init__(self, bundle_engine, services):
        self.bundle_engine = bundle_engine
        self.service_handles = [ServiceHandle(bundle_engine, x) for x in services]

    def start_services_without_prerequisites(self):
        for service_handle in self.service_handles:
            if service_handle.no_prerequisites():
                service_handle.start_setup()

    def _setup_finished(self):
        return [x for x in self.service_handles if x.setup_finished]

    def pipes(self):
        stdout_pipes = [x.stdout_pipe for x in self._setup_finished()]
        stderr_pipes = [x.stderr_pipe for x in self._setup_finished()]
        return stdout_pipes + stderr_pipes


    def _not_started(self):
        return [x for x in self.service_handles if not x.started]

    def _ready_services(self):
        return [x.service for x in self.service_handles if x.ready]

    def not_ready_services(self):
        return [x.service for x in self.service_handles if not x.ready]


    def all_services_ready(self):
        for service_handle in self.service_handles:
            if not service_handle.ready:
                return False
        return True

    def longest_service_name(self):
        """Length of the longest service name."""
        return max([len(service_handle.service.name) for service_handle in self.service_handles] + [0])

    def tail_setup_and_poststart(self):
        for service_handle in self.service_handles:
            name = service_handle.service.name
            self.bundle_engine.initiate_tail(name.lower() + "_setup.out", "Setup {}".format(name), None)
            self.bundle_engine.initiate_tail(name.lower() + "_setup.err", "Err Setup {}".format(name), Fore.YELLOW)

            self.bundle_engine.initiate_tail(name.lower() + "_poststart.out", "Post {}".format(name), None)
            self.bundle_engine.initiate_tail(name.lower() + "_poststart.err", "Err Post {}".format(name), Fore.YELLOW)

    def _get_service_from_pipe(self, pipe):
        for service_handle in self.service_handles:
            if pipe == service_handle.stdout_pipe:
                return (service_handle, False)
            if pipe == service_handle.stderr_pipe:
                return (service_handle, True)

    def handle_input(self, pipe_handle, data):
        service_handle, is_error = self._get_service_from_pipe(pipe_handle)
        lines = [line for line in data.decode("utf8", "ignore").split('\n') if line != ""]

        for line in lines:
            decoded_line = line

            if is_error:
                self.bundle_engine.writeline("Err {0}".format(service_handle.service.name), decoded_line, color=Fore.YELLOW)
            else:
                self.bundle_engine.writeline("    {0}".format(service_handle.service.name), decoded_line)

            if not service_handle.loaded:
                if service_handle.log_line_readiness_checker(line):
                    service_handle.loaded = True
                    self.bundle_engine.logline("{0} Loaded.".format(service_handle.service.name))
                    service_handle.poststart_run()
                    service_handle.poststart_started = True

    def poll(self):
        for service_handle in self.service_handles:
            # Start the process if its setup thread is done
            if not service_handle.setup_finished and hasattr(service_handle, 'setup_runner') and not service_handle.setup_runner.is_alive():
                service_handle.setup_finished = True
                service_handle.start_process()
                if service_handle.stdout_pipe is not None:
                    service_handle.stdout_pipe.start_read(self.bundle_engine.on_pipe_read)
                if service_handle.stderr_pipe is not None:
                    service_handle.stderr_pipe.start_read(self.bundle_engine.on_pipe_read)

            # If a poststart thread is done...
            if service_handle.poststart_started and not service_handle.ready and not service_handle.poststart_runner.is_alive():
                service_handle.ready = True

                # Start any services which rely upon this one as a prerequisite
                for notstartedservice in self._not_started():
                    if set(notstartedservice.service.needs).issubset(set(self._ready_services())):
                        notstartedservice.start_setup()

            # Throw exception if a service suddenly stops.
            if service_handle.process_started and service_handle.is_dead():
                self.bundle_engine.messages_to_driver.put(
                    ServiceSuddenStopException(
                        "Service '{}' stopped suddenly.".format(
                            service_handle.service.name
                        )
                    )
                )

    def stop(self):
        for service_handle in self.service_handles:
            service_handle.stop()

        still_service_engines = list(self.service_handles)

        for i in range(0, int(self.bundle_engine.service_bundle.shutdown_timeout * 100)):
            for service_handle in still_service_engines:
                if service_handle.is_dead():
                    still_service_engines.remove(service_handle)
                    self.bundle_engine.logline("{0} Stopped".format(service_handle.service.name))

            if len(still_service_engines) == 0:
                break

            time.sleep(0.01)

        # If after timeout seconds there are services remaining, commit service genocide.
        for service_handle in still_service_engines:
            self.bundle_engine.warnline("Killing {0}".format(service_handle.service.name))
            if not service_handle.is_dead():
                service_handle.kill()



from hitchserve.hitch_dir import HitchDir
from hitchserve.service_bundle import ServiceBundle
from hitchserve.service_handle import ServiceHandle
from hitchserve.test_engine import TestEngine
from hitchserve.hitch_service import Service
from hitchserve.hitch_exception import HitchException
from hitchserve.hitch_exception import HitchServeException
from hitchserve.hitch_exception import ServiceStartupTimeoutException
from hitchserve.hitch_exception import ServiceSuddenStopException



import os
import io

def _write(handle, message):
    if isinstance(handle, io.TextIOWrapper):
        handle.write(message)
    else:
        handle.write(message.encode('utf8'))
    handle.flush()

def log(message):
    """Output to stdout."""
    import sys
    _write(sys.stdout, message)

def warn(message):
    """Output to stderr."""
    import sys
    _write(sys.stderr, message)



from hitchserve.service_handle import ServiceHandle
from hitchserve.service_engine import ServiceEngine
from hitchserve.hitch_exception import ServiceStartupTimeoutException
from colorama import Fore, Back, Style
import functools
import termios
import signal
import psutil
import time
import pyuv
import sys
import os

# TODO: Report actual directory being used at the beginning of the logs.
# TODO: Move all from-previous-test-run-process killing code here.
# TODO: Rename TestEngine -> OrchestrationEngine

class TestEngine(object):
    """Engine (and thread) that brings together all I/O and events driving hitchserve."""
    _timedout = False
    _ready = False
    _driver_sent_shutdown_signal = False
    _shutdown_triggered = False
    ipython_on = False

    def __init__(self, service_bundle):
        self.service_bundle = service_bundle

    def signal_cb(self, handle, signum):
        """Handle ctrl-C if not in ipython shell. Always shutdown on SIGTERM."""
        SIGNALS_TO_NAMES_DICT = dict((getattr(signal, n), n) \
            for n in dir(signal) if n.startswith('SIG') and '_' not in n )
        self.logline("Signal {} received...".format(SIGNALS_TO_NAMES_DICT.get(signum)))
        if signum in (signal.SIGTERM, signal.SIGHUP, signal.SIGQUIT, signal.SIGINT):
            self.stop()

    def handle_tail(self, fullfilename, title, color, handle, filename, events, error):
        with open(fullfilename, 'r') as filehandle:
            filehandle.seek(self.tail_positions[fullfilename])
            tailportion = filehandle.read()
        self.tail_positions[fullfilename] = self.tail_positions[fullfilename] + len(tailportion) #os.stat(fullfilename).st_size

        for line in tailportion.split('\n'):
            if line != "":
                if color is None:
                    self.logline(line, title=title)
                else:
                    self.logline(line, title=title, color=color)


    # TODO: Close tail handles after set up / poststart is finished.
    def initiate_tail(self, stdlog, title, color):
        filename = self.service_bundle.hitch_dir.stdlogdir() + os.sep + stdlog
        with open(filename, "a"):
            os.utime(filename, None)
        self.tail_positions[filename] = 0
        tail = pyuv.fs.FSEvent(self.loop)
        tail.start(filename, 0, functools.partial(self.handle_tail, filename, title, color))
        self.tail_handles.append(tail)
        with open(filename, "a"):
            os.utime(filename, None)

    def start(self, messages_to_driver, messages_to_bundle_engine):
        """Orchestrate processes and I/O pipes."""

        os.setpgrp()
        self.service_bundle.hitch_dir.save_pgid("bundleengine", os.getpgid(os.getpid()))
        self.service_engine = ServiceEngine(self, self.service_bundle.values())

        self.start_time = time.time()
        self.messages_to_driver = messages_to_driver
        self.messages_to_bundle_engine = messages_to_bundle_engine

        os.chdir(self.service_bundle.hitch_dir.project_directory)
        self.logfile = open(self.service_bundle.hitch_dir.testlog(), "a")

        self.loop = pyuv.Loop.default_loop()

        self.pipe_logfile = pyuv.Pipe(self.loop)
        self.pipe_logfile.open(self.logfile.fileno())

        if not self.service_bundle.quiet:
            self.pipe_stdout = pyuv.Pipe(self.loop)
            self.pipe_stdout.open(sys.stdout.fileno())

        self.service_engine.start_services_without_prerequisites()


        self.timer_handler = pyuv.Timer(self.loop)
        self.timer_handler.start(self.poll_handler, 0.01, 0.01)

        self.tail_handles = []
        self.tail_positions = {}

        self.initiate_tail("driver.out", "Test", None)
        self.initiate_tail("driver.err", "Err Test", Fore.YELLOW)

        self.service_engine.tail_setup_and_poststart()

        self.signal_h = pyuv.Signal(self.loop)
        self.signal_h.start(self.signal_cb, signal.SIGINT)
        self.signal_h.start(self.signal_cb, signal.SIGTERM)

        self.loop.run()
        self.loop = None
        os.kill(os.getpid(), signal.SIGKILL)

    def _close_pipes(self):
        """Close all the pipes in order to shut the engine down."""
        if not self.service_bundle.quiet:
            if not self.pipe_stdout.closed:
                self.pipe_stdout.close()
        for pipe in self.service_engine.pipes():
            if pipe is not None and not pipe.closed:
                pipe.close()
        if not self.signal_h.closed:
            self.signal_h.close()
        if not self.timer_handler.closed:
            self.timer_handler.close()
        for handle in self.tail_handles:
            handle.close()
        self.logfile.close()

    def writeline(self, identifier, line, color=''):
        """Log a line to the log file and/or stdout."""
        reset_all = Fore.RESET + Back.RESET + Style.RESET_ALL
        full_line = "{0}{1}[{2}] {3}{4}\n".format(
            reset_all,
            color,
            identifier.rjust(self.service_engine.longest_service_name() + 10),
            line,
            reset_all
        )
        if not self.ipython_on and not self.service_bundle.quiet:
            self.pipe_stdout.write(full_line.encode('utf-8'))
        self.pipe_logfile.write(full_line.encode('utf-8'))

    def logline(self, line, title="Hitch", color=''):
        self.writeline(title, line, color)

    def warnline(self, line):
        self.writeline("WARNING", line, color=Fore.YELLOW)

    def on_pipe_read(self, pipe_handle, data, error):
        if data is None:
            pass
        else:
            self.service_engine.handle_input(pipe_handle, data)

    def poll_handler(self, timer_handle):
        """Handle messages from the test thread and timeout."""
        self.service_engine.poll()

        if not self._ready and self.service_engine.all_services_ready():
            startup_duration = time.time() - self.start_time
            self.logline("READY in {0:.1f} seconds.".format(startup_duration), color=Style.BRIGHT)
            self.messages_to_driver.put("READY")
            self._ready = True

        if not self.messages_to_bundle_engine.empty():
            msg = self.messages_to_bundle_engine.get()
            if not self._driver_sent_shutdown_signal and msg == "SHUTDOWN":
                self._driver_sent_shutdown_signal = True
                self.stop()
            if msg == "IPYTHONON":
                self.ipython_on = True
            elif msg == "IPYTHONOFF":
                self.ipython_on = False

        if not self._ready and not self._timedout:
            if time.time() - self.start_time > self.service_bundle.startup_timeout:
                self.warnline("TIMEOUT")
                self._timedout = True
                self.messages_to_driver.put(
                    ServiceStartupTimeoutException(
                        "Services not ready after {} second timeout: {}".format(
                            self.service_bundle.startup_timeout,
                            ", ".join([x.name for x in self.service_engine.not_ready_services()]),
                        )
                    )
                )


    def stop(self):
        """Shut down all hitchserve services and processes started from driver process, kill if necessary."""
        if not self._shutdown_triggered:
            self._shutdown_triggered = True
            self.ipython_on = False

            start_shutdown_time = time.time()

            for child in psutil.Process(os.getppid()).children():
                if child is not None and child.is_running() and child.pid != os.getpid():
                    for grandchild in psutil.Process(child.pid).children(recursive=True):
                        try:
                            grandchild.send_signal(signal.SIGINT)
                            self.logline("Stopping PID {} : {}".format(grandchild.pid, " ".join(grandchild.cmdline())))
                        except psutil.NoSuchProcess:
                            pass
                    try:
                        child.send_signal(signal.SIGINT)
                        self.logline("Stopping PID {} : {}".format(child.pid, " ".join(child.cmdline())))
                    except psutil.NoSuchProcess:
                        pass

            self.service_engine.stop()
            still_alive = False

            if time.time() < start_shutdown_time + self.service_bundle.shutdown_timeout:
                sleep_time = self.service_bundle.shutdown_timeout - (time.time() - start_shutdown_time)
                for i in range(0, int(sleep_time * 100)):
                    still_alive = False
                    for child in psutil.Process(os.getppid()).children():
                        if child is not None and child.is_running() and child.pid != os.getpid():
                            for grandchild in psutil.Process(child.pid).children(recursive=True):
                                try:
                                    if grandchild.status != "zombie":
                                        still_alive = True
                                except psutil.NoSuchProcess:
                                    pass
                    if not still_alive:
                        break
                    time.sleep(0.01)

            if still_alive:
                for child in psutil.Process(os.getppid()).children():
                    if child is not None and child.is_running() and child.pid != os.getpid():
                        for grandchild in psutil.Process(child.pid).children(recursive=True):
                            try:
                                fullname = " ".join(grandchild.cmdline())
                                grandchild.send_signal(signal.SIGKILL)
                                self.logline("Killing PID {} : {}".format(grandchild.pid, grandchild.name))
                            except psutil.NoSuchProcess:
                                pass
                        try:
                            child.send_signal(signal.SIGKILL)
                            self.logline("Killing PID {} : {}".format(child.pid, child.name()))
                        except psutil.NoSuchProcess:
                            pass

            self.logline("Shutdown successful in {0:.1f} seconds".format(time.time() - start_shutdown_time))
            self._close_pipes()






from hitchserve.hitch_exception import ServiceMisconfiguration
from hitchserve.hitch_dir import HitchDir
from hitchserve import service_logs
from hitchserve.utils import log, warn
import multiprocessing
import subprocess
import faketime
import inspect
import signal
import psutil
import time
import pyuv
import sys
import os
import re


class HitchServiceException(Exception):
    """Exception associated with a HitchServe service."""
    pass


# TODO: Allow stopping and starting and *waiting* of services from other thread using queue.

class Subcommand(object):
    """Command associated with a service.

       Examples include Postgres's psql or pg_dump, 'redis-cli' or Django's manage."""

    def __init__(self, *args, **kwargs):
        """Define a subcommand.

        Args:
            *args (str): Sequence of program arguments needed to run the command.
            directory (Optional[str]): Directory the command is run in.
            env_vars (Optional[dict]): Environment variable to feed to the subcommand.
        """
        self.command = list(args)
        self.directory = kwargs['directory'] if 'directory' in kwargs else None
        self.env_vars = kwargs['env_vars'] if 'env_vars' in kwargs else None

    @property
    def directory(self):
        """str: Directory subcommand runs in."""
        return self._directory

    @directory.setter
    def directory(self, value):
        self._directory = value

    def run(self, shell=False, ignore_errors=False, stdin=False, check_output=False):
        """Run subcommand.

        Args:
            shell (Optional[bool]): Run command using shell (default False)
            ignore_errors (Optional[bool]): If the command has a non-zero return code, don't raise an exception (default False)
            stdin (Optional[bool]): Plug input from stdin when running command (default False)
            check_output (Optional[bool]): Return command output as string (default False)

        Returns:
            String if check_output is True, else None.

        Raises:
            subprocess.CalledProcessError when the command has an error, unless ignore_errors is True.
        """
        os.chdir(self.directory)
        try:
            kwargs = {
                'stderr': sys.stderr,
                'stdin': sys.stdin if stdin else None,
                'env': self.env_vars,
                'shell': shell,
            }
            if check_output:
                return subprocess.check_output(self.command, **kwargs).decode("utf8")
            else:
                kwargs['stdout'] = sys.stdout
                return subprocess.check_call(self.command, **kwargs)
        except subprocess.CalledProcessError:
            if ignore_errors:
                pass
            else:
                raise

class Service(object):
    """A process to run, monitored and interacted with for the duration of a test."""
    stop_signal = signal.SIGINT

    def __init__(self, command=None, log_line_ready_checker=None, directory=None, no_libfaketime=False, env_vars=None, needs=None):
        """Define and configure a service.

        Each service has a command, directory and function which checks
        each line of the logs to ascertain readiness.

        Args:
            command (List(str)): Sequence of program arguments needed to run the service.
            log_line_ready_checker (Function[str]): Function which returns True when passed a line which indicates service readiness.
            directory (Optional[str]): Directory the service command is run in. Defaults to project directory specified in service bundle.
            no_libfaketime (Optional[bool]): If True, don't run service with libfaketime. Useful if libfaketime breaks the service.
            env_vars (Optional[dict]): Dictionary of environment variables to feed to the service when running it.
            needs (Optional[List[Service]]): List of services which must be started before this service will run.

        Raises:
            ServiceMisconfiguration when the wrong parameters are passed.
        """
        self.no_libfaketime = no_libfaketime
        self.directory = directory
        self.command = command
        self.env_vars = {} if env_vars is None else env_vars
        self.needs = needs
        self.log_line_ready_checker = log_line_ready_checker
        self._pid = multiprocessing.Value('i', 0)

        if not inspect.isfunction(log_line_ready_checker):
            raise ServiceMisconfiguration(
                "log_line_ready_checker must be a function that takes one argument: a string"
            )

        if len(inspect.getargspec(log_line_ready_checker).args) != 1:
            raise ServiceMisconfiguration(
                "log_line_ready_checker must take only one argument."
            )

        if needs is not None:
            if type(needs) != list:
                raise ServiceMisconfiguration(
                    "needs must be of type List[Service] or None. You did not pass a list!"
                )

            for need in needs:
                if not isinstance(need, Service):
                    raise ServiceMisconfiguration(
                        "needs must be of type List[Service] or None."
                    )

    def setup(self):
        """Method that is run before starting the service."""
        pass

    def poststart(self):
        """Method that is run immediately after detecting that the service has started."""
        pass

    @property
    def pid(self):
        """int: UNIX process id of the service process."""
        return self._pid.value

    @pid.setter
    def pid(self, value):
        self._pid.value = value

    @property
    def process(self):
        """psutil.Process: psutil Process object of the service."""
        return psutil.Process(self.pid)

    @property
    def name(self):
        """str: Service name."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.logs = service_logs.ServiceLogs(value)

    @property
    def env_vars(self):
        """dict: All environment variables fed to the service."""
        if not self.no_libfaketime:
            faketime_filename = self.service_group.hitch_dir.faketime()

            env_vars = dict(os.environ)
            env_vars.update(self._env_vars)
            env_vars.update(faketime.get_environment_vars(faketime_filename))
        else:
            env_vars = dict(os.environ)
            env_vars.update(self._env_vars)
        return env_vars

    @env_vars.setter
    def env_vars(self, value):
        self._env_vars = value

    @property
    def directory(self):
        """str: Directory that the service is run in."""
        if self._directory is None:
            return self.service_group.hitch_dir.hitch_dir
        else:
            return self._directory

    @directory.setter
    def directory(self, value):
        self._directory = value

    @property
    def command(self):
        """List[str]: Command used to run the service."""
        return self._command

    @command.setter
    def command(self, value):
        self._command = value

    def log(self, line):
        log("{}\n".format(line))

    def warn(self, line):
        warn("{}\n".format(line))

    def subcommand(self, *args):
        """Get subcommand acting on a service. Subcommand will run in service directory
           and with the environment variables used to run the service itself.

        Args:
            *args: Arguments to run command (e.g. "redis-cli", "-n", "1")

        Returns:
            Subcommand object.
        """
        return Subcommand(*args, directory=self.directory, env_vars=self.env_vars)



# -*- coding: utf-8 -*
from setuptools.command.install import install
from setuptools import find_packages
from setuptools import setup
import subprocess
import codecs
import sys
import os

def read(*parts):
    # intentionally *not* adding an encoding option to open
    # see here: https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(os.path.join(os.path.abspath(os.path.dirname(__file__)), *parts), 'r').read()

setup(name="hitchserve",
      version="0.4.6",
      description="Service orchestration library for the Hitch testing framework.",
      long_description=read('README.rst'),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
          'Topic :: Software Development :: Quality Assurance',
          'Topic :: Software Development :: Testing',
          'Topic :: Software Development :: Libraries',
          'Operating System :: Unix',
          'Environment :: Console',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
      ],
      keywords='hitch testing framework bdd tdd declarative tests testing epoll kqueue service',
      author='Colm O\'Connor',
      author_email='colm.oconnor.github@gmail.com',
      url='https://github.com/hitchtest/hitchserve/',
      license='AGPL',
      install_requires=['humanize>=0.5.1', 'colorama', 'psutil', 'pyuv', 'tblib', 'faketime', ],
      packages=find_packages(exclude=[]),
      package_data={},
      zip_safe=False,
      include_package_data=True,
)



from hitchserve.hitch_dir import HitchDir
from hitchserve.hitch_exception import BundleMisconfiguration
from hitchserve.hitch_exception import HitchServeException
from hitchserve.hitch_service import Service
from hitchserve.test_engine import TestEngine
from hitchserve.utils import log, warn
from datetime import datetime as python_datetime
from datetime import timedelta as python_timedelta
import faketime
import multiprocessing
import colorama
import subprocess
import humanize
import inspect
import termios
import psutil
import signal
import time
import six
import sys
import os


# TODO : Allow stopping, starting, restarting and waiting until stop/start is finished.

class ServiceBundle(object):
    def __init__(self, project_directory, startup_timeout=15.0, shutdown_timeout=5.0, hitch_dir=None, quiet=False):
        self._shutdown_initiated = False
        self.aborted = False
        self._services = {}
        self.quiet = quiet
        self.timedelta = python_timedelta(0)
        self.project_directory = os.path.abspath(project_directory)
        self.hitch_dir = HitchDir(self.project_directory, hitch_dir=hitch_dir)
        self.startup_timeout = startup_timeout
        self.shutdown_timeout = shutdown_timeout

    def __len__(self):
        return len(self._services.items())

    def __getitem__(self, key):
        if key not in self._services:
            raise IndexError
        return self._services[key]

    def __setitem__(self, key, value):
        if not isinstance(value, Service):
            raise BundleMisconfiguration("'{}' must be of type 'Service'".format(key))

        if "[" in key or "]" in key:
            raise BundleMisconfiguration("[ and ] characters not allowed in service name")

        if " " in key:
            raise BundleMisconfiguration("Spaces are not allowed in service names.")

        if key in ["Test", "Hitch", "Harness", ]:
            raise BundleMisconfiguration("{0} is not an allowed service name.".format(key))
        self._services[key] = value
        self._services[key].name = key
        self._services[key].service_group = self

        if type(self._services[key].command) is not list:
            raise BundleMisconfiguration(
                "'{}' command should be a list of strings - e.g. ['startservice', 'arg1', 'arg2']".format(key)
            )

        if len([x for x in self._services[key].command if type(x) is str]) < len(self._services[key].command):
            raise BundleMisconfiguration(
                "'{}''s command should be a list of strings. Your list contains non-strings.".format(key)
            )

        if self._services[key].command is None:
            raise BundleMisconfiguration("'{}' command must be a list, not {}".format(key, self._services[key].command))

        if subprocess.call(["which", self._services[key].command[0]], stdout=subprocess.PIPE) != 0:
            raise BundleMisconfiguration("'{}' command must call an existing file, not '{}'".format(key, self._services[key].command[0]))

        if not os.path.exists(self._services[key].directory):
            raise BundleMisconfiguration("'{}' directory '{}' must exist.".format(key, self._services[key].directory))

    def __repr__(self):
        return str(self._services)

    def items(self):
        return self._services.items()

    def values(self):
        return self._services.values()

    def keys(self):
        return self._services.keys()

    def __iter__(self):
        return self._services.__iter__()


    def log(self, message):
        """Print a normal priority message."""
        log(message)

    def warn(self, message):
        """Print a higher priority message."""
        warn(message)

    def check_pid(self, pid):
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def _shutdown_old_processes(self):
        old_pid = self.hitch_dir.old_pid()
        if old_pid is not None:
            if self.check_pid(old_pid):
                log("Shutting down existing test...\n")
                old_process = psutil.Process(old_pid)

                for child in old_process.children(recursive=True):
                    try:
                        child.send_signal(signal.SIGTERM)
                        child.send_signal(signal.SIGQUIT)
                        log("Stopping {} (PID {}) : '{}'\n".format(child.name(), child.pid, ' '.join(child.cmdline())))
                    except psutil.NoSuchProcess:
                        pass

                try:
                    old_process.wait(self.shutdown_timeout)
                    return
                except psutil.TimeoutExpired:
                    pass

            old_pgids = self.hitch_dir.old_pgids()
            if len(old_pgids) > 0:
                log("Killing off left-over processes from previous test...\n")
                processes = [p for p in psutil.process_iter() if os.getpgid(p.pid) in old_pgids]

                for process in processes:
                    try:
                        process.send_signal(signal.SIGKILL)
                        warn("Killing {} (PID {}) : '{}'\n".format(process.name(), process.pid, ' '.join(process.cmdline())))
                    except psutil.NoSuchProcess:
                        pass


    def startup(self, interactive=True):
        self._shutdown_old_processes()
        if hasattr(sys.stdin, 'fileno'):
            try:
                self._orig_stdin_termios = termios.tcgetattr(sys.stdin.fileno())
            except termios.error as e:
                # Inappropriate ioctl for device -- quiet mode
                if e.args[0] == 25:
                    self._orig_stdin_termios = None
                else:
                    raise
            self._orig_stdin_fileno = sys.stdin.fileno()
        else:
            self._orig_stdin_termios = None
            self._orig_stdin_fileno = None

        for service in self.values():
            service.logs.set_logfilename(self.hitch_dir.testlog())


        self.hitch_dir.clean()

        self.hitch_dir.save_pgid("driver_process", os.getpgid(os.getpid()))
        self.hitch_dir.save_pid(os.getpid())

        self.messages_to_driver = multiprocessing.Queue()
        self.messages_to_bundle_engine = multiprocessing.Queue()

        service_process_args = (self.messages_to_driver, self.messages_to_bundle_engine)

        self.bundle_engine = TestEngine(self)
        self._service_process = multiprocessing.Process(target=self.bundle_engine.start, args=service_process_args)
        self._service_process.start()

        self.redirect_stdout()

        what_happened = self.messages_to_driver.get()

        if what_happened == "READY":
            if interactive:
                self.start_interactive_mode()
            return
        elif type(what_happened) == tuple and len(what_happened) == 3:
            if interactive:
                self.start_interactive_mode()
            six.reraise(*what_happened)
        elif isinstance(what_happened, Exception):
            if interactive:
                self.start_interactive_mode()
            raise what_happened
        else:
            if interactive:
                self.start_interactive_mode()
            raise HitchServeException("Unknown exception occurred")

    def redirect_stdout(self):
        """Redirect stdout to file so that it can be tailed and aggregated with the other logs."""
        self.hijacked_stdout = sys.stdout
        self.hijacked_stderr = sys.stderr
        # 0 must be set as the buffer, otherwise lines won't get logged in time.
        sys.stdout = open(self.hitch_dir.driverout(), "ab", 0)
        sys.stderr = open(self.hitch_dir.drivererr(), "ab", 0)

    def unredirect_stdout(self):
        """Redirect stdout and stderr back to screen."""
        if hasattr(self, 'hijacked_stdout') and hasattr(self, 'hijacked_stderr'):
            sys.stdout = self.hijacked_stdout
            sys.stderr = self.hijacked_stderr

    def start_interactive_mode(self):
        if hasattr(self, 'messages_to_bundle_engine'):
            # Tell bundle engine to stop logging
            self.messages_to_bundle_engine.put("IPYTHONON")

            # Ensure that IPYTHONON message to bundle engine is received (poll handler run every 0.01 secs)
            time.sleep(0.05)

            # Reset any terminal color changes
            log("{}{}{}".format(colorama.Fore.RESET, colorama.Back.RESET, colorama.Style.RESET_ALL))
            warn("{}{}{}".format(colorama.Fore.RESET, colorama.Back.RESET, colorama.Style.RESET_ALL))

            # Stop redirecting test stdout/err to file.
            self.unredirect_stdout()

            # If termios attr of stdin is None, don't do anything to stdin
            if self._orig_stdin_termios is not None:
                # Make stdin blocking - so that redis-cli (among others) can work.
                import fcntl
                flags = fcntl.fcntl(self._orig_stdin_fileno, fcntl.F_GETFL)
                if flags & os.O_NONBLOCK:
                    fcntl.fcntl(self._orig_stdin_fileno, fcntl.F_SETFL, flags & ~os.O_NONBLOCK)

                # Set termios attributes in a sensible state so that the terminal can be used interactively
                termios.tcsetattr(self._orig_stdin_fileno, termios.TCSANOW, self._orig_stdin_termios)
            else:
                warn("\n\n==========> Can't start interactive mode when stdin is redirected to /dev/null\n")

    def stop_interactive_mode(self):
        if hasattr(self, 'messages_to_bundle_engine'):
            # Make stdin non-blocking again
            import fcntl
            flags = fcntl.fcntl(sys.stdin.fileno(), fcntl.F_GETFL)
            if flags & ~os.O_NONBLOCK:
                fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK)

            # Start redirecting test stdout/err to file again, so it can be logged along with other services
            self.redirect_stdout()

            # Tell bundle engine to start logging again
            self.messages_to_bundle_engine.put("IPYTHONOFF")

            # Ensure that IPYTHONOFF message to bundle engine is received (poll handler run every 0.01 secs)
            time.sleep(0.05)

    def time_travel(self, datetime=None, timedelta=None, seconds=0, minutes=0, hours=0, days=0):
        """Mock moving forward or backward in time by shifting the system clock fed to the services tested.

        Note that all of these arguments can be used together, individually or not at all. The time
        traveled to will be the sum of all specified time deltas from datetime. If no datetime is specified,
        the deltas will be added to the current time.

        Args:
            datetime (Optional[datetime]): Time travel to specific datetime.
            timedelta (Optional[timedelta]): Time travel to 'timedelta' from now.
            seconds (Optional[number]): Time travel 'seconds' seconds from now.
            minutes (Optional[number]): Time travel 'minutes' minutes from now.
            hours (Optional[number]): Time travel 'hours' hours from now.
            days (Optional[number]): Time travel 'days' days from now.
        """
        if datetime is not None:
            self.timedelta = datetime - python_datetime.now()
        if timedelta is not None:
            self.timedelta = self.timedelta + timedelta
        self.timedelta = self.timedelta + python_timedelta(seconds=seconds)
        self.timedelta = self.timedelta + python_timedelta(minutes=minutes)
        self.timedelta = self.timedelta + python_timedelta(hours=hours)
        self.timedelta = self.timedelta + python_timedelta(days=days)
        log("Time traveling to {}\n".format(humanize.naturaltime(self.now())))
        faketime.change_time(self.hitch_dir.faketime(), self.now())

    def now(self):
        """Get a current (mocked) datetime. This will be the current datetime unless you have time traveled."""
        return python_datetime.now() + self.timedelta

    def pstree(self):
        os.system("pstree -panl {}".format(os.getppid()))

    def connect_to_ipykernel(self, service_name, timeout=10):
        """Connect to an IPython kernel as soon as its message is logged."""
        kernel_line = self._services[service_name].logs.tail.until(
            lambda line: "--existing" in line[1], timeout=10, lines_back=5
        )
        kernel_json_file = kernel_line.replace("--existing", "").strip()
        self.start_interactive_mode()
        subprocess.check_call([
            sys.executable, "-m", "IPython", "console", "--existing", kernel_json_file
        ])
        self.stop_interactive_mode()

    def shutdown(self):
        if not self._shutdown_initiated:
            self._shutdown_initiated = True
            if hasattr(self, '_service_process'):
                if self._service_process.is_alive():
                    self.messages_to_bundle_engine.put("SHUTDOWN")
                    try:
                        psutil.Process(self._service_process.pid).wait()
                    except psutil.NoSuchProcess:
                        pass

            self.unredirect_stdout()
            self.hitch_dir.remove_run_dir()
            log("{}{}{}".format(colorama.Fore.RESET, colorama.Back.RESET, colorama.Style.RESET_ALL))
            warn("{}{}{}".format(colorama.Fore.RESET, colorama.Back.RESET, colorama.Style.RESET_ALL))
            if self.aborted:
                warn("ABORT\n")



from colorama import Fore, Back, Style
from tblib import pickling_support
import multiprocessing
import subprocess
import traceback
import functools
import inspect
import signal
import psutil
import time
import pyuv
import sys
import os


class ServiceHandle(object):
    """This is the handle that starts, stops the services and hooks it up to the service and bundle engines."""
    def __init__(self, bundle_engine, service):
        self.bundle_engine = bundle_engine
        self.service = service
        self.started = False
        self.setup_finished = False
        self.process_started = False
        self.poststart_started = False
        self.poststart_finished = False
        self.ready = False
        self.loaded = False
        self.process = None
        self.stdout_pipe = None
        self.stderr_pipe = None

        if self.service.log_line_ready_checker:
            self.log_line_readiness_checker = self.service.log_line_ready_checker
        else:
            self.log_line_readiness_checker = lambda line: False

    def start_setup(self):
        def run_setup():
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            try:
                sys.stdout = open(self.bundle_engine.service_bundle.hitch_dir.setup_out(self.service.name), "ab", 0)
                sys.stderr = open(self.bundle_engine.service_bundle.hitch_dir.setup_err(self.service.name), "ab", 0)
                self.service.setup()
            except Exception as e:
                pickling_support.install()
                self.bundle_engine.messages_to_driver.put(sys.exc_info())

        self.started = True
        self.setup_runner = multiprocessing.Process(target=run_setup)
        self.setup_runner.start()

    def start_process(self):
        self.bundle_engine.logline("Starting {}".format(self.service.name))
        self.bundle_engine.logline("Directory: {}".format(self.service.directory))
        self.bundle_engine.logline("Command: {}".format(' '.join(self.service.command)))
        os.chdir(self.service.directory)

        try:
            self.process = subprocess.Popen(
                self.service.command,
                bufsize=0,                  # Ensures that all stdout/err is pushed to us immediately.
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=self.service.env_vars,
                preexec_fn=os.setpgrp       # Ctrl-C signal is not passed on to the process.
            )
            self.service.pid = self.process.pid
            self.process_started = True
        except Exception as e:
            pickling_support.install()
            self.bundle_engine.messages_to_driver.put(sys.exc_info())
            return

        self.bundle_engine.service_bundle.hitch_dir.save_pgid(self.service.name.lower(), os.getpgid(self.process.pid))

        self.stdout_pipe = pyuv.Pipe(self.bundle_engine.loop)
        self.stdout_pipe.open(self.process.stdout.fileno())

        self.stderr_pipe = pyuv.Pipe(self.bundle_engine.loop)
        self.stderr_pipe.open(self.process.stderr.fileno())


    def poststart_run(self):
        def run_poststart():
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            try:
                sys.stdout = open(self.bundle_engine.service_bundle.hitch_dir.poststart(self.service.name), "ab", 0)
                sys.stderr = open(self.bundle_engine.service_bundle.hitch_dir.poststart_err(self.service.name), "ab", 0)
                self.service.poststart()
            except Exception as e:
                pickling_support.install()
                self.bundle_engine.messages_to_driver.put(sys.exc_info())
                return

        self.poststart_runner = multiprocessing.Process(target=run_poststart)
        self.poststart_runner.start()

    def stop(self):
        """Ask politely, first, with SIGINT and SIGQUIT."""
        if hasattr(self, 'process'):
            if self.process is not None:
                try:
                    is_running = self.process.poll() is None
                except AttributeError:
                    is_running = False

                if is_running:
                    self.bundle_engine.logline("Stopping {0}".format(self.service.name))
                    self.term_signal_sent = True

                    # Politely ask all child processes to die first
                    try:
                        for childproc in psutil.Process(self.process.pid).children(recursive=True):
                            childproc.send_signal(signal.SIGINT)
                    except psutil.NoSuchProcess:
                        pass
                    except AttributeError:
                        pass

                    try:
                        self.process.send_signal(self.service.stop_signal)
                    except OSError as e:
                        if e.errno == 3:        # No such process
                            pass
                else:
                    self.bundle_engine.warnline("{0} stopped prematurely.".format(self.service.name))
            else:
                self.bundle_engine.warnline("{0} stopped prematurely.".format(self.service.name))
        else:
            self.bundle_engine.warnline("{0} was never successfully started.".format(self.service.name))

    def is_dead(self):
        if not hasattr(self, 'process'):
            return True
        else:
            if self.process is None:
                return True
            else:
                return self.process.poll() is not None or psutil.Process(self.process.pid).status() == 'zombie'

    def kill(self):
        """Murder the children of this service in front of it, and then murder the service itself."""
        if not self.is_dead():
            self.bundle_engine.warnline("{0} did not shut down cleanly, killing.".format(self.service.name))
            try:
                if hasattr(self.process, 'pid'):
                    for child in psutil.Process(self.process.pid).children(recursive=True):
                        os.kill(child.pid, signal.SIGKILL)
                    self.process.kill()
            except psutil.NoSuchProcess:
                pass

    def no_prerequisites(self):
        return self.service.needs is None or self.service.needs == []
